#import "AppDelegate.h"
#import <SafariServices/SafariServices.h>

@interface AppDelegate ()

@property (weak) IBOutlet NSWindow *window;
@end

@implementation AppDelegate {
	NSRunningApplication *_safari;
}

static CFStringRef const KeepsWindowsOpenPref = CFSTR("NSQuitAlwaysKeepsWindows");
static NSString *const SafariBundleID = @"com.apple.Safari";
static void *TerminateSafariContext = &TerminateSafariContext;

- (void)copySafari {
	NSString *safariPath = [[NSWorkspace sharedWorkspace] absolutePathForAppBundleWithIdentifier:SafariBundleID];
	if (safariPath == nil) {
		NSLog(@"Safari not found!");
		return;
	}
	NSArray<NSString *> *appSupport = NSSearchPathForDirectoriesInDomains(NSApplicationSupportDirectory, NSUserDomainMask, YES);
	if ([appSupport count] == 0) {
		NSLog(@"Application Support not found!");
		return;
	}
	NSString *myAppSupport = [appSupport[0] stringByAppendingPathComponent:@"com.lapcatsoftware.SafariPrivacyTest"];
	[[NSFileManager defaultManager] removeItemAtPath:myAppSupport error:NULL]; // Ignore errors
	NSString *mySafari = [myAppSupport stringByAppendingPathComponent:@"Safari.app"];
	NSError *error = nil;
	if (![[NSFileManager defaultManager] createDirectoryAtPath:mySafari withIntermediateDirectories:YES attributes:nil error:&error]) {
		NSLog(@"Error creating my Safari: %@", error);
		return;
	}
	// Only copy the Contents, because there's a rootless attribute on Safari.app,
	// which causes a copy error.
	NSString *mySafariContents = [mySafari stringByAppendingPathComponent:@"Contents"];
	if (![[NSFileManager defaultManager] copyItemAtPath:[safariPath stringByAppendingPathComponent:@"Contents"] toPath:mySafariContents error:&error]) {
		NSLog(@"Error copying Safari: %@", error);
		return;
	}
	
	NSString *jsPath = [[NSBundle mainBundle] pathForResource:@"HTMLViewController" ofType:@"js"];
	if (jsPath == nil) {
		NSLog(@"HTMLViewController.js not found!");
		return;
	}
	NSString *replacementPath = [[mySafariContents stringByAppendingPathComponent:@"Resources"] stringByAppendingPathComponent:@"HTMLViewController.js"];
	if (![[NSFileManager defaultManager] removeItemAtPath:replacementPath error:&error]) {
		NSLog(@"Error removing HTMLViewController.js: %@", error);
		return;
	}
	if (![[NSFileManager defaultManager] copyItemAtPath:jsPath toPath:replacementPath error:&error]) {
		NSLog(@"Error copying HTMLViewController.js: %@", error);
		return;
	}
	if (![[NSWorkspace sharedWorkspace] launchApplication:mySafari]) {
		NSLog(@"Could not launch Safari!");
		return;
	}
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context {
	if (context == TerminateSafariContext) {
		[self copySafari];
	} else {
		[super observeValueForKeyPath:keyPath ofObject:object change:change context:context];
	}
}

- (void)terminateSafari {
	NSArray<NSRunningApplication *> *safaris = [NSRunningApplication runningApplicationsWithBundleIdentifier:SafariBundleID];
	if ([safaris count] != 1) {
		NSLog(@"Unexpected Safaris: %@", safaris);
		return;
	}
	NSRunningApplication *safari = safaris[0];
	if ([safari terminate]) {
		[self copySafari];
	} else {
		NSLog(@"Observing Safari termination");
		_safari = safari;
		[_safari addObserver:self forKeyPath:@"terminated" options:0 context:TerminateSafariContext];
	}
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification {
	CFPreferencesSetValue(KeepsWindowsOpenPref, kCFBooleanTrue, kCFPreferencesAnyApplication, kCFPreferencesCurrentUser, kCFPreferencesCurrentHost);
	CFPreferencesSynchronize(kCFPreferencesAnyApplication, kCFPreferencesCurrentUser, kCFPreferencesCurrentHost);
	[SFSafariApplication showPreferencesForExtensionWithIdentifier:@"com.lapcatsoftware.SafariPrivacyTest.SafariPrivacyTestExtension" completionHandler:^(NSError * _Nullable error) {
		if (error != nil)
			NSLog(@"Error: %@", error);
		else
			dispatch_after(dispatch_time(DISPATCH_TIME_NOW, (int64_t)(2.0 * NSEC_PER_SEC)), dispatch_get_main_queue(), ^{
				[self terminateSafari];
			});
	}];
}

- (void)applicationWillTerminate:(NSNotification *)aNotification {
	CFPreferencesSetValue(KeepsWindowsOpenPref, kCFBooleanFalse, kCFPreferencesAnyApplication, kCFPreferencesCurrentUser, kCFPreferencesCurrentHost);
	CFPreferencesSynchronize(kCFPreferencesAnyApplication, kCFPreferencesCurrentUser, kCFPreferencesCurrentHost);
}

@end
